/*
 * NoICE Map File Generator - version 3.2
 *
 * This program reads a DDS listing file, and generates a .MAP file
 * containing information about addresses, file names, line numbers,
 * and global symbols.
 *
 * Copyright 1994-2000 Dave Dunfield
 * Freely distributable.
 *
 * JLH: Modified 30-Apr-98 by John Hartman for NoICE.
 * JLH: (Starting from Dunfield verion of 2-23-98)
 * JLH: Changes are bracketed or followed by "JLH" comments
 * JLH: Modified 26-May-1998 to change help text (version 3.1)
 * JLH: Modified 31-Aug-1998 to change help text (version 3.2)
 *
 * Permission is granted to modify and distribute this program without
 * payment or royalty, as long as the original copyright notices are not
 * removed.
 *
 * Tool vendors are encouraged to modify this program to generate MAP files
 * compatible with their tools, and distribute it to their customers. If you
 * do so, please add indication of the company doing the modification, and
 * MAP format supported. I would also appreciate your sending me a copy of
 * the modified program for posting on our BBS.
 *
 * I have provided large "box" comments showing the main areas where you will
 * have to perform modification in order to support a certain MAP file format.
 * Note that these code blocks can be re-arranged, enhanced or eliminated to
 * perform output in the order and format which your MAP file requires.
 * The structure of this program is for guideline only.
 *
 ** Portability notes are identified by **
 * Conditionals using _MICROC_ are provided which allow compilation with
 * either Micro-C/PC 3.14, or Turbo-C 2.0 - You may need to modify these
 * areas for use with a different compiler.
 *
 * Compiled by John Hartman using the venerable Zortech V3.0r2 using
 *    ztc -2 -a1 -bx -C -mli -pac -R -s -u -w mapnoi
 * A compiled version (MAPNOI.EXE) is included with the NoICE distribution,
 * available from  http://ourworld.compuserve.com/homepages/jhartman/
 */
#include <stdio.h>

/*JLH: Functions and data for NoICE */
static void SetProcessorType( char *typeString );
static void ProcessorHeader();
static void ArgumentOut( unsigned allocationSize, unsigned offset );
static void LocalVarOut( unsigned allocationSize, unsigned offset );
static void SetFile( int fileNumber );
static void SetFunction( int functionNumber );
static char* AddressOut( unsigned segment, unsigned address );
static char* get_string2(unsigned i);

int usingPages = 0;
unsigned char page[10];
int processorType = -1;
/*end JLH*/


/* Misc. MAPNOI parameters */
#define	LINE_SIZE	250		/* Maximum size of input line */
#define	MAX_FILES	75		/* Maximum number of files to handle */
#define	MAX_SYMB	5000		/* Maximum number of symbols */
#define	MAX_DIMS	500		/* Maximum number of array dimensions */

/* Offsets into listing lines */
#define	OPCODE	7			/* Opcode bytes begin here */
#define	SOURCE	32			/* Source text begins here */

/*
 * Bits found in the "type" entry of symbol table, also
 * used on the expression stack to keep track of element types,
 * and passed to the code generation routines to identify types.
 *
 * The POINTER definition indicates the level of pointer indirection,
 * and should contain all otherwise unused bits, in the least significant
 * (rightmost) bits of the word.
 *
 * MAPNOI reuses the REFERENCE bit to indicate a LOCAL symbol.
 */
#define REFERENCE	0x8000		/* symbol has been referenced */
#define INITED		0x4000		/* symbol is initialized */
#define ARGUMENT	0x2000		/* symbol is a function argument */
#define EXTERNAL	0x1000		/* external attribute */
#define STATIC		0x0800		/* static attribute */
#define REGISTER	0x0400		/* register attribute */
#define	TVOID		0x0200		/* void attribute */
#define UNSIGNED	0x0100		/* unsigned attribute */
#define BYTE		0x0080		/* 0=16 bit, 1=8 bit */
#define	SYMTYPE		0x0070		/* symbol type (see below) */
#define ARRAY		0x0008		/* symbol is an array */
#define POINTER		0x0007		/* level of pointer indirection */

/*
 * Symbol type designator bits (SYMTYPE field above)
 */
#define	VARIABLE	0x00		/* Symbol is a simple variable */
#define	MEMBER		0x10		/* Symbol is a structure member */
#define	STRUCTURE	0x20		/* Symbol is a structure template */
#define	FUNCTION	0x30		/* Symbol is a function */
#define	GOTOLAB		0x40		/* Symbol is a "goto" label */

/*
 * Filename table	- tracks files occuring in the listing
 */
unsigned fcount = 0,			/* Count of filenames */
	current_file = 0,			/* File currently active */
	filename[MAX_FILES];		/* File name (offset) */

/*
 * Symbol table		- tracks symbols occuring in the listing
 */
unsigned scount,				/* Count of symbols */
	name[MAX_SYMB],				/* Symbol name (offset) */
	type[MAX_SYMB],				/* Symbol type */
	addr[MAX_SYMB],				/* Symbol address/offset */
	value1[MAX_SYMB],			/* F:StkAlloc V:Segment */
	value2[MAX_SYMB];			/* F:EndAddr  V:Array=DimIdx Struct=Size*/
unsigned char
	file[MAX_SYMB];				/* Index to source file name */

/*
 * Dimension table	- tracks dimensions of symbols with array type
 */
unsigned dim_pool[MAX_DIMS], dcount = 0;

/*
 * Command line argument switches
 */
char amap = -1, cmap = -1, quiet = 0, ucase = 0, dup = 0, equ = 0, list = 0;

/*
 * Misc. global variables
 */
unsigned string_seg,			/* External string segment */
	mapseg = 0,					/* Current "map" segment in file */
	lst_line = 0,				/* Current line number in listing */
	acount = 0;					/* Count of address/line records */
char buffer[LINE_SIZE+1];		/* Line input buffer */
FILE *fpi, *fpo;				/* Input & Output file pointers */


#ifndef _MICROC_

/*JLH: using far array, since my compiler doesn't have allocmem et al.*/
#include <stdlib.h>
char* stringMem;
int allocmem(unsigned paragraphs, unsigned *pSeg)
{
	/* Hack: my calloc can't allocate quite 64K.  Back off til it works */
	stringMem = (char*)calloc( paragraphs-2, 16 );
	return (stringMem != NULL) ? -1 : 0;
}
char peek( unsigned seg, unsigned offset)
{
	return stringMem[offset];
}
void poke(unsigned seg, unsigned offset, char value)
{
	stringMem[offset] = value;
}
/*end JLH*/

/*
 ** If your compiler puts '\n' on the end of lines read by fgets()
 ** (this will be evidenced by newlines after every filename in the
 ** output file)... Use this version
 */
static char *fgets(char *dest, int size, FILE *fp)
{
	register int c;
	register char *ptr;

	ptr = dest;
	while(--size) {
		if((c = getc(fp)) < 0) {
			if(ptr == dest)
				dest = 0;
			break; }
		if(c == '\n')
			break;
		*ptr++ = c; }
	*ptr = 0;
	return dest;
}
#endif

/*
 ** Save a string into the external pool
 */
static unsigned string_top = 0; /*JLH - visible at file scope for diagnostics*/
static unsigned save_string(char *name)
{
	unsigned p;

	p = string_top;
	do
		poke(string_seg, string_top++, *name);
	while(*name++);
	return p;
}

/*
 ** Retrieve a string from the external pool
 */
static char *get_string(unsigned i)
{
	char *ptr;
	static char temp[65];

	ptr = temp;		/* Point to our copy */
	while(*ptr++ = peek(string_seg, i++));
	return temp;
}

/*
 * Test for uppercase alphabetic
 */
static isupper(char c)
{
	return (c >= 'A') && (c <= 'Z');
}

/*
 * Test for space character
 */
static isspace(char c)
{
	return (c == ' ') || (c == '\t');
}

/*
 * Test for numeric digit
 */
static isdigit(char c)
{
	return (c >= '0') && (c <= '9');
}

/*
 * Test for a valid symbol name
 */
static issymbol(char *ptr)
{
	char c;
	c = *ptr;
	do {
		if( ((c < 'a') || (c > 'z'))
		&&	((c < 'A') || (c > 'Z'))
		&&	((c < '0') || (c > '9'))
		&&	(c != '_') )
			return 0; }
	while(!isspace(c = *++ptr));
	return 1;
}

/*
 * Test for string1 beginning with string2
 */
static strbeg(char *str1, char *str2)
{
	while(*str2)
		if(*str1++ != *str2++)
			return 0;
	return 1;
}

/*
 * Get a number in specified base with digit limit
 */
static get_number(char *ptr, unsigned base, unsigned digits)
{
	unsigned value, c;

	value = 0;
	while(isspace(*ptr))
		++ptr;
	do {
		if(isdigit(c = *ptr))			/* convert numeric digits */
			c -= '0';
		else if(c >= 'a')				/* convert lower case alphabetics */
			c -= ('a' - 10);
		else if(c >= 'A')				/* convert upper case alphabetics */
			c -= ('A' - 10);
		else							/* not a valid "digit" */
			break;
		if(c >= base)					/* outside of base */
			break;
		value = (value * base) + c;		/* include in total */
		++ptr; }
	while(--digits);					/* enforce maximum digits */
	return value;
}

/*
 * Copy string up to a delimiter
 */
static char *copy_parse(char *dest, char *source)
{
	while(isspace(*source))
		++source;
	if(ucase) {
		while(*source && !isspace(*source))
			*dest++ = toupper(*source++); }
	else {
		while(*source && !isspace(*source))
			*dest++ = *source++; }
	*dest = 0;
	return source;
}

/*
 * Find a global symbol
 */
static unsigned find_global(char *s)
{
	unsigned i;
	i = 0;
	while(i < scount) {
		if(!strcmp(s, get_string(name[i])))				/* found! */
			return i+1;
		if((type[i]&SYMTYPE) == FUNCTION) {		/* Function - skip locals */
			while(type[++i] & REFERENCE);
			continue; }
		++i; }
	return 0;
}

/*
 * First pass... Define all files and symbols
 */
static first_pass()
{
	unsigned i, j;
	char *ptr;
	char temp[65], temp1[10];
	static unsigned last_func;

top:
	if(!fgets(buffer, LINE_SIZE, fpi))			/* End of input */
		return;
	if(*buffer<' ') goto top;					/* Blank line */
	if(strbeg(buffer, "DUNFIELD")) goto top;	/* Header line */
	if(strbeg(buffer, "SYMBOL")) return;		/* End of listing */

	if(*(buffer+SOURCE) == '*') {				/* Comment line */
		if(isdigit(*(ptr=buffer+(SOURCE+1)))) {	/* 'C' line number */
			goto top; }
		if(*ptr != '#') {						/* ASM comment */
			current_file = 0;
			goto top; }
		if(strbeg(++ptr, "file ")) {			/* C file name */
			copy_parse(temp, buffer+(SOURCE+7));
			for(current_file=0; current_file < fcount; ++current_file)
				if(!strcmp(get_string(filename[current_file]), ptr))
					goto top;
			filename[fcount++] = save_string(temp);
			goto top; }
		if(strbeg(ptr, "cpu ")) {			/* JLH: processor type */
			SetProcessorType(ptr+4);
			goto top; }
		if(strbeg(ptr, "link")) {			/* Linked file */
			current_file = 0;
			goto top; }
		if(strbeg(ptr, "map")) {			/* Map segment */
			mapseg = atoi(ptr+3);
			goto top; }
		if(strbeg(ptr, "fun")) {			/* Function beginning */
			ptr = copy_parse(temp, ptr+3);
			name[scount] = save_string(temp);
			file[scount] = current_file;
			ptr = copy_parse(temp1, ptr);
			type[last_func = scount] = FUNCTION;
			value1[scount++] = atoi(temp1);
			goto top; }
		if(strbeg(ptr, "end")) {			/* Function ending */
			value2[last_func] = get_number(buffer, 16, 4) - 1;
			goto top; }
		if(strbeg(ptr, "lcl")) {			/* Local variable */
			ptr = copy_parse(temp, ptr+3);
			ptr = copy_parse(temp1, ptr);
			i = atoi(temp1) | REFERENCE;
			if((i & SYMTYPE) == MEMBER)
				*temp |= 0x80;
			name[j = scount] = save_string(temp);
			ptr = copy_parse(temp1, ptr);
			value2[scount] = addr[scount] = atoi(temp1);
			++scount;
			goto setsym; }
		if(strbeg(ptr, "gbl")) {				/* Global variable */
			ptr = copy_parse(temp, ptr+3);
			ptr = copy_parse(temp1, ptr);
			i = atoi(temp1) & ~REFERENCE;
			if((i & SYMTYPE) == MEMBER)
				*temp |= 0x80;
			ptr = copy_parse(temp1, ptr);
			value2[scount] = atoi(temp1);
			if(i & EXTERNAL)					/* Do not handle extern */
				goto top;
			if(j = find_global(temp))
				--j;
			else
				name[j = scount++] = save_string(temp);
		setsym:
			file[j] = current_file;
			if((type[j] = i) & ARRAY) {
				ptr = copy_parse(temp1, ptr);
				value2[j] = dcount;
				dim_pool[dcount++] = i = atoi(temp1);
				while(i--) {
					ptr = copy_parse(temp1, ptr);
					dim_pool[dcount++] = atoi(temp1); } }
			goto top; } }

	/*
	 * Global symbol - make/amend symbol table entry
	 */
	if(issymbol(ptr = buffer+SOURCE)) {
		copy_parse(temp, ptr);
		if(j = find_global(temp)) {			/* Already defined */
			addr[--j] = get_number(buffer, 16, 4);
			if((type[j] & SYMTYPE) != FUNCTION)
				value1[j] = mapseg; }
		else {
			name[scount] = save_string(temp);
			file[scount] = current_file;	/* New symbol */
			addr[scount] = get_number(buffer, 16, 4);
			value1[scount] = mapseg;
			type[scount++] = SYMTYPE; /* Unknown */ } }

	/*
	 * Set addresses for static symbols
	 */
	if(*ptr == '?') {		/* Possible static */
		ptr = buffer+(SOURCE+1);
		if(isupper(*ptr) && isupper(ptr[1]) && isdigit(ptr[2])) {
			ptr += 3;
			while(isdigit(*ptr)) ++ptr;
			if(issymbol(ptr)) {				/* Looks like a static */
				copy_parse(temp, buffer+SOURCE);
				for(i=0; i < scount; ++i)
					if(!strcmp(get_string(name[i]), temp)) {
						addr[i] = get_number(buffer, 16, 4);
						if((type[i] & SYMTYPE) != FUNCTION)
							value1[i] = mapseg;
						break; } } } }
	goto top;
}

/*
 * Second pass... Generate address/line-number records
 */
static dump_address()
{
	char *ptr;
	char temp[50], cflag;
	unsigned a, l, cline, last_cline;

	mapseg = current_file = cflag = last_cline = 0;

top:
	if(!fgets(buffer, LINE_SIZE, fpi))			/* End of input */
		return;
	++lst_line;
	if(*buffer<' ') goto top;					/* Blank line */
	if(strbeg(buffer, "DUNFIELD")) goto top;	/* Header line */
	if(strbeg(buffer, "SYMBOL")) return;		/* End of listing */

	if(*(buffer+SOURCE) == '*') {				/* Comment line */
		if(isdigit(*(ptr=buffer+(SOURCE+1)))) {	/* 'C' line number */
			cline = get_number(buffer+(SOURCE+1), 10, 0);
			cflag = cmap;
			goto top; }
		if(*ptr != '#') {						/* ASM comment */
			current_file = 0;
			goto top; }
		if(strbeg(++ptr, "file ")) {			/* C file name */
			copy_parse(temp, buffer+(SOURCE+7));
			for(current_file=0; current_file < fcount; ++current_file)
				if(!strcmp(get_string(filename[current_file]), temp))
					break;
			goto top; }
		if(strbeg(ptr, "link")) {				/* Linked file */
			current_file = 0;
			goto top; }
		if(strbeg(ptr, "map")) {				/* Map segment */
			mapseg = atoi(ptr+3);
			goto top; }
		goto top; }

	if(*(buffer+OPCODE) == ' ')
		goto top;

	if(current_file) {
		if(!cflag)
			goto top;
		if((cline == last_cline) && !dup)
			goto top;
		last_cline = l = cline; }
	else {
		if(!amap)
			goto top;
		l = lst_line; }
	a = get_number(buffer, 16, 4);
	if(list)
		printf("@%04x: F:%03u L:%u\n", a, current_file, l);
	++acount;

/**************************************************/
/* Dump an address/line record to the output file */
/**************************************************/
	SetFile(current_file);
	fprintf(fpo, "LINE %u %s\n", l, AddressOut(mapseg,a));

	cflag = 0;
	goto top;
}

/************************************************/
/* Dump a symbol definitions to the output file */
/* JLH: this version replaces original function */
/************************************************/
static dump_symbols()
{
	int i, j, k, t, st, func;
	char *ptr;
	static int structSize = 0;

	for(i=0; i < scount; ++i) {
		ptr = get_string(name[i]);
		*ptr &= 0x7f;
		st = (t = type[i]) & SYMTYPE;
		if(list)
			printf("S%-4u: F:%03u T:%04x A:%04x 1:%04x 2:%04x %s\n",
				i, file[i], t, addr[i], value1[i], value2[i], ptr);
		if((st == SYMTYPE) && equ)		/* Do not dump unknown's */
			continue;
		if(*ptr == '?') {			/* Trim prefix on static symbol */
			ptr += 3;
			while(isdigit(*ptr)) ++ptr; }

		if (structSize && ((t & SYMTYPE) != MEMBER)) {
			/* end of structure definition */
			fprintf(fpo, "ENDS 0x%X\n", structSize );
			structSize = 0; }

		SetFile( file[i] );

		switch(st) {
			case STRUCTURE :
				fprintf(fpo, "STRUCT 0 %s\n", ptr );
				structSize = value2[i];
				continue;
			case FUNCTION :
				func = i;
				SetFunction(-1);	/* close any open function defore DEFS */
				if (t & STATIC)
					fprintf(fpo, "DEFS %s %s %%code\n", ptr, AddressOut( 0, addr[i]));
				else
					fprintf(fpo, "DEF %s %s %%code\n", ptr, AddressOut( 0, addr[i]));
				SetFunction(func);		/* open function definition */
				continue;
			case MEMBER :
				/* Since MC doesn't tell datatype of structure
				 * variables, define a global symbol with
				 * name equal to structure element.  User can
				 * do home-made offset into "struct" byte array:
				 * W foo.es_i won't work, but
				 * W foo[es_i] %int will */
				fprintf(fpo, "DEF %s 0:0x%X", ptr, value2[i] );

				/* Structure member
				 * MCC does not emit a STRUCT entry for anonymous structures.
				 * Thus, don't emit DEFTYPEs for them either */
				if (structSize) fprintf(fpo, "\nDEFTYPE 0x%X %s", value2[i], ptr );
				break;
			case VARIABLE :
			case SYMTYPE :
				if(t & REFERENCE) {
					SetFunction( func );
					fprintf(fpo, "DEFS %s ", ptr);
					if (t & STATIC)
						fprintf(fpo, "%s", AddressOut(value1[i],addr[i]));
					else {
						if (t & ARGUMENT)
							/* function argument on stack */
							ArgumentOut(value1[func], addr[i]);
						else
							/* local variable on stack */
							LocalVarOut(value1[func], addr[i]); }}
				else {
					SetFunction(-1);	/* symbol NOT in a function */
					if (t & STATIC)
						fprintf(fpo, "DEFS %s %s", ptr, AddressOut(value1[i],addr[i]));
					else
						fprintf(fpo, "DEF %s %s", ptr, AddressOut(value1[i],addr[i])); }
				break;
			case GOTOLAB : ; }

		/* show type */
		fprintf(fpo, " %%" );
		j = (t & POINTER);
		while (j--)
			fprintf(fpo, "*" );

		if (t & TVOID)
			fprintf(fpo, "void" );
		else {
			fprintf(fpo, "%c%s", (t & UNSIGNED) ? 'U' : 'S',
				(t & BYTE) ? "08" : "16" ); }


		if(t & ARRAY) {
			j = value2[i];		/* Get dim table index */
			k = dim_pool[j++];	/* Get # dimensions */
			while(k--)
				fprintf(fpo, "[%u]", dim_pool[j++]); }
		putc('\n', fpo); }
}
/*end JLH*/

/*
 * Main program
 */
main(int argc, char *argv[])
{
	unsigned i;
	char *ptr, c;

	for(i=1; i < argc; ++i) {
		if(*(ptr = argv[i]) == '-') {	/* Enable switch */
			while(*++ptr) switch(toupper(*ptr)) {
				case 'A' : amap = 0;	break;
				case 'C' : cmap = 0;	break;
				case 'D' : dup = -1;	break;
				case 'E' : equ = -1;	break;
				case 'L' : list = -1;	break;
				case 'Q' : quiet = -1;	break;
				case 'U' : ucase = -1;	break;

				/*JLH: Processor type is rest of argument */
				case 'P' :
					SetProcessorType(ptr+1);
					ptr = ".";	/* end while loop */
					break;

				/*JLH: Segment mappings */
				case '0' :
				case '1' :
				case '2' :
				case '3' :
				case '4' :
				case '5' :
				case '6' :
				case '7' :
				case '8' :
				case '9' :
					page[*ptr - '0'] = get_number(ptr+1, (unsigned)16, (unsigned)2);
					usingPages = 1;
					ptr = ".";	/* end while loop */
					break;

                case '?':
                    printf("MAPNOI processes .LST files output by DDS C\n" );
                    printf("Specify input file as 'file' or 'file.LST'.  Output is 'file.NOI\n");
                    printf("\nMAPNOI file [-acdelqu] [-Pnnnn] [-0nn]\n" );
                    printf("Switches are set automatically by DDS C-startup file, or manually:\n");
                    printf(" -P6809   target is 6809\n");
                    printf(" -P6811   target is 68HC11               -P6812   target is 68HC12\n");
                    printf(" -P8051T  target is 8051, Tiny model     -P8051S  target is 8051, Small model\n");
                    printf(" -P8051C  target is 8051, Compact model  -P8051M  target is 8051, Medium model\n");
                    printf(" -P8051L  target is 8051, Large model\n");
                    printf(" -P8096   target is 8096\n");
                    printf(" -0nn     Set page for segment 0 to 'nn'. Default is 0\n");
                    printf("  ...      ...\n");
                    printf(" -9nn     Set page for segment 9 to 'nn'. Default is 0\n");
                    exit(0);

				default: goto badopt; }
			continue; }
        else {
			copy_parse(ptr = buffer, argv[1]);
            continue; }
	badopt:
		printf("Invalid option: %s\n", argv[i]);
		exit(-1); }

/*JLH*/
	if(!quiet)
		printf("MAPNOI version 3.2: Process Dunfield Micro-C list files for NoICE.\n");

	if(argc < 2) {
        printf("  Derived by John Hartman from MAPGEN V2.0 to support NoICE\n");
        printf("  MAPGEN Copyright 1994-2000 Dave Dunfield.  Freely distributable.\n");
        printf("\nFor DDS compiler 'CCxx', and a single file 'file.c', run\n");
        printf("  CCxx file -CPLS\n");
		printf("  MAPNOI file\n");
        printf("\nFor multiple files 'file1.c' ... 'filex.c', run\n");
        printf("  CCxx file1 -CAPS\n");
        printf("  ...\n");
        printf("  CCxx filex -CAPS\n");
        printf("  SLINK file1.asm ... filex.asm -S file.asm\n");
        printf("  ASMxx file -CFS\n");
		printf("  MAPNOI file\n" );
		printf("\nAfter MAPNOI, run NoICExx, and give the command 'PLAY file.noi'\n" );
		printf("This will give you source line and symbol information.");
		printf("\nFor a complete list of MAPNOI command line switches, run\n");
		printf("  MAPNOI -?\n" );
		exit(-1); }
/* end JLH*/

/*
 ** Allocate a 64K external segment for string storage
 */
#ifdef	_MICROC_
	if(!(string_seg = alloc_seg(4096))) {
#else
	if(allocmem(4096, &string_seg) != -1) {
#endif
	printf("Out of memory\n");
	exit(-1); }

	/* Parse filename & extension from passed argument */
	/* copy_parse(ptr = buffer, argv[1]); */
	while(c = *ptr) {
		if(c == '.')
			goto noext;
		++ptr; }
	copy_parse(ptr, ".LST");
noext:

	/* Parse filenames */
	filename[fcount++] = save_string(buffer);
	if(!(fpi = fopen(buffer, "r"))) {
		printf("Cannot open input file: %s\n", buffer);
		exit(-2); }
	copy_parse(ptr, ".NOI");	/*JLH: change extension to NoICE preferred*/
	if(!(fpo = fopen(buffer, "w"))) {
		printf("Cannot open output file: %s\n", buffer);
		exit(-2); }

	/*JLH: output preamble here, since we need the file name in buffer*/
	fprintf(fpo, "CASE 2\n" );
	copy_parse(ptr, ".HEX");
	fprintf(fpo, "LOAD %s\n", buffer );
	fprintf(fpo, "DEFT 0 char %%S08\n" );
	fprintf(fpo, "DEFT 0 int %%S16\n" );
	/*end JLH*/

	first_pass();		/* Evaluate all symbols */

	rewind(fpi);

	/*JLH: output any processor-specific info (Allows first_pass to change
	 * processor type, overriding command line, if desired)
	 */
	ProcessorHeader();

	/* Dump symbol definitions to the output file */
	dump_symbols();

	/* Dump address records to the output file */
	dump_address();

	SetFile( -1 );	/*JLH: final ENDFILE */

	fclose(fpo);
	fclose(fpi);

	if(!quiet) {
		printf("%u files, %u symbols, %u dimensions, %u address entries\n",
			fcount, scount, dcount, acount);
		printf("%u string bytes\n", string_top); } /*JLH: added second printf*/
}

/* JLH:  Remainder of file is NoICE-specific code */

#define PRO_6809  0
#define PRO_6811  1
#define PRO_6812  2
#define PRO_8051T 3
#define PRO_8051S 4
#define PRO_8051C 5
#define PRO_8051M 6
#define PRO_8051L 7
#define PRO_8096  8

static char* processorName[] = {
	"6809",
	"6811",
	"6812",
	"8051T",
	"8051 TINY",
	"8051S",
	"8051 SMALL",
	"8051C",
	"8051 COMPACT",
	"8051M",
	"8051 MEDIUM",
	"8051L",
	"8051 LARGE",
	"8096",
	NULL };

/* PARALLEL ARRAY to processorName (or make them a struct) */
static int processorTypeCode[] = {
	PRO_6809,	/* 6809 */
	PRO_6811,	/* 6811 */
	PRO_6812,	/* 6812 */
	PRO_8051T,	/* 8051T */
	PRO_8051T,	/* 8051 Tiny */
	PRO_8051S,	/* 8051S */
	PRO_8051S,	/* 8051 Small */
	PRO_8051C,	/* 8051C */
	PRO_8051C,	/* 8051 Compact */
	PRO_8051M,	/* 8051M */
	PRO_8051M,	/* 8051 Medium */
	PRO_8051L,	/* 8051L */
	PRO_8051L,	/* 8051 Large */
	PRO_8096	/* 8096 */
	};

/* Set processor type:	affects frame pointers etc */
void SetProcessorType( char *typeString )
{
	int i;

	strupr(typeString);
	for (i=0; (processorName[i] != NULL); i++) {
		if (!strcmp(processorName[i],typeString)) {
			processorType = processorTypeCode[i];
			return; }}
	printf("\nUnknown processor type %s", typeString );
	exit(-2);
}

/* generate processor-specific header commands */
void ProcessorHeader()
{
	switch (processorType) {
	case PRO_8051T:
	case PRO_8051M:
		usingPages = 1;
		page[0] = 0x0C;
		page[1] = 0x0C;
		page[2] = 0x0;
		page[3] = 0x0D;
	case PRO_8051S:
		fprintf(fpo, "MSBFIRST 0\n" );
		break;

	case PRO_8051L:
		usingPages = 1;
		page[0] = 0x0C;
		page[1] = 0x0C;
		page[2] = 0x0;
		page[3] = 0x0D;
	case PRO_8051C:
		fprintf(fpo, "MSBFIRST 0\n" );
		fprintf(fpo, "DEFREG ?stack 0:8 2\n" );
		break;

	case PRO_8096:
		fprintf(fpo, "DEFREG ?BP 0x1A 2\n" );
		fprintf(fpo, "DEFREG ?AX 0x1C 2\n" );
		fprintf(fpo, "DEFREG ?BX 0x22 2\n" );
		fprintf(fpo, "DEFREG ?CX 0x20 2\n" );
		fprintf(fpo, "DEFREG ?DX 0x1E 2\n" );
		fprintf(fpo, "DEFREG ?SI 0x24 2\n" );
		fprintf(fpo, "DEFREG ?DI 0x26 2\n" );
		break; }
}

/* emit value of argument relative to appropriate frame pointer */
void ArgumentOut( unsigned allocationSize, unsigned offset )
{
	switch (processorType) {
	case PRO_6809:
	case PRO_6812:
		fprintf(fpo, "SP+0x%X", 2+allocationSize + offset); break;
	case PRO_6811:
		fprintf(fpo, "SP+0x%X", 3+allocationSize + offset); break;

	case PRO_8051T:
	case PRO_8051S:
	case PRO_8051M:
		fprintf(fpo, "0:SP-0x%X", 3+allocationSize + offset); break;

	case PRO_8051C:
	case PRO_8051L:
		fprintf(fpo, "0:SP-0x%X", 3+offset); break;

	case PRO_8096:
		fprintf(fpo, "?BP+0x%X", 4+offset); break;

	default:
		printf("\nProcessor type not specified" );
        exit(2);}
}

/* emit value of local variable relative to appropriate frame pointer */
void LocalVarOut( unsigned allocationSize, unsigned offset )
{
	switch (processorType) {
	case PRO_6809:
	case PRO_6812:
		fprintf(fpo, "SP+0x%X", offset); break;
	case PRO_6811:
		fprintf(fpo, "SP+0x%X", 1+offset); break;

	case PRO_8051T:
	case PRO_8051S:
	case PRO_8051M:
		fprintf(fpo, "0:SP-0x%X", allocationSize - 1 - offset); break;

	case PRO_8051C:
	case PRO_8051L:
		fprintf(fpo, "D:?stack+0x%X", offset); break;

	case PRO_8096:
		fprintf(fpo, "?BP-0x%X", allocationSize-offset); break;

	default:
		printf("\nProcessor type not specified" );
        exit(2);}
}

/* If file number changes, output a FILE command */
void SetFile( int fileNumber )
{
	static int lastFileSeen = -1;

	if (fileNumber != lastFileSeen) {
		if (lastFileSeen >= 0) {
			SetFunction(-1);
			fprintf(fpo, "ENDFILE\n" ); }

		if (file >= 0)
			fprintf(fpo, "FILE %s\n", get_string2(filename[fileNumber]));

		lastFileSeen = fileNumber; }
}

/* If function number changes, output a FUNCTION command */
void SetFunction( int function )
{
	static int lastFunctionSeen = -1;
	char *ptr;

	if (function != lastFunctionSeen) {
		if (lastFunctionSeen >= 0)
			fprintf(fpo, "ENDF %s\n", AddressOut(0, value2[lastFunctionSeen]));

		if (function >= 0) {
			ptr = get_string2(name[function]);
		if(*ptr == '?') {
			ptr += 3;
			while(isdigit(*ptr)) ++ptr;
			fprintf(fpo, "SFUNC %s %s\n", ptr, AddressOut( 0, addr[function])); }
		else
			fprintf(fpo, "FUNC %s %s\n", ptr, AddressOut( 0, addr[function])); }

		lastFunctionSeen = function; }
}

/* return pointer to paged address as ASCII */
char* AddressOut( unsigned segment, unsigned address )
{
	static char buf[ 10 ];
	char *ptr = buf;
	if (usingPages)
		ptr += sprintf(buf, "%02X:", page[ segment ] );

	sprintf(ptr, "0x%X", address );
	return buf;
}

/*
 ** Retrieve a string from the external pool
 *
 * Second copy, since static buffer is used
 */
static char *get_string2(unsigned i)
{
	char *ptr;
	static char temp[65];

	ptr = temp;		/* Point to our copy */
	while(*ptr++ = peek(string_seg, i++));
	return temp;
}
/* end JLH */
